#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Generic, TypeVar

from pydantic import BaseModel

from argilla.server.daos.models.records import BaseAnnotationDB, BaseRecordDB, BaseRecordInDB


class ServiceBaseAnnotation(BaseAnnotationDB):
    pass


class BulkResponse(BaseModel):
    dataset: str
    processed: int
    failed: int = 0


ServiceAnnotation = TypeVar("ServiceAnnotation", bound=ServiceBaseAnnotation)


class ServiceBaseRecordInputs(BaseRecordInDB[ServiceAnnotation], Generic[ServiceAnnotation]):
    pass


class ServiceBaseRecord(BaseRecordDB[ServiceAnnotation], Generic[ServiceAnnotation]):
    pass


ServiceRecord = TypeVar("ServiceRecord", bound=ServiceBaseRecord)
