(window.webpackJsonp=window.webpackJsonp||[]).push([[193,194],{1254:function(e,t,n){"use strict";n.r(t),n.d(t,"useDatasetSettingViewModel",(function(){return o}));var a=n(29),r=(n(42),n(945)),c=n(975),o=function useDatasetSettingViewModel(){var e=Object(c.useDatasetViewModel)(),t=Object(r.a)((function(){return[].concat(Object(a.a)(e.breadcrumbs.value),[{link:null,name:"settings"}])}));return Object(r.b)((function(){e.loadDataset()})),Object.assign(Object.assign({},e),{breadcrumbs:t})}},945:function(e,t,n){"use strict";var a=n(0);n.d(t,"a",(function(){return a.computed})),n.d(t,"b",(function(){return a.onBeforeMount})),n.d(t,"c",(function(){return a.onMounted})),n.d(t,"d",(function(){return a.onUnmounted}));a.default,a.default.util.warn},975:function(e,t,n){"use strict";n.r(t),n.d(t,"useDatasetViewModel",(function(){return i}));n(21),n(9),n(27);var a=n(186),r=n(55),c=n(39),o=n(284),u=n(279),s=n(283),__awaiter=function(e,t,n,a){return new(n||(n=Promise))((function(r,c){function fulfilled(e){try{step(a.next(e))}catch(e){c(e)}}function rejected(e){try{step(a.throw(e))}catch(e){c(e)}}function step(e){e.done?r(e.value):function adopt(e){return e instanceof n?e:new n((function(t){t(e)}))}(e.value).then(fulfilled,rejected)}step((a=a.apply(e,t||[])).next())}))},i=function useDatasetViewModel(){var e=Object(a.c)(!1),t=Object(a.g)(),n=Object(a.f)().value.params.id,i=Object(s.a)().state,d=Object(r.useResolve)(o.a),f=Object(a.a)((function(){return p(i)})),l=function handleError(e){var t="";switch(e){case u.a.ERROR_FETCHING_DATASET_INFO:t="Can't get dataset info for dataset_id: ".concat(n);break;case u.a.ERROR_FETCHING_WORKSPACE_INFO:t="Can't get workspace info for dataset_id: ".concat(n);break;default:t="There was an error on fetching dataset info and workspace info. Please try again"}var a={message:t,numberOfChars:t.length,type:"error"};c.a.dispatch("notify",a)},p=function createBreadcrumbs(e){return[{link:{name:"datasets"},name:"Home"},{link:{path:"/datasets?workspace=".concat(e.workspace)},name:e.workspace},{link:{name:null},name:e.name}]};return{dataset:i,datasetId:n,isLoadingDataset:e,loadDataset:function loadDataset(){return __awaiter(void 0,void 0,void 0,regeneratorRuntime.mark((function _callee(){return regeneratorRuntime.wrap((function _callee$(a){for(;;)switch(a.prev=a.next){case 0:return a.prev=0,e.value=!0,a.next=4,d.execute(n);case 4:a.next=10;break;case 6:a.prev=6,a.t0=a.catch(0),l(a.t0.response),t.push("/");case 10:return a.prev=10,e.value=!1,a.finish(10);case 13:case"end":return a.stop()}}),_callee,null,[[0,6,10,13]])})))},breadcrumbs:f}}}}]);