#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from argilla.server.daos.backend.mappings.helpers import mappings


def text2text_mappings():
    """Text2Text index mappings"""
    return {
        "_source": mappings.source(
            excludes=[
                "predicted",
                "predicted_as",
                "predicted_by",
                "annotated_as",
                "annotated_by",
                "score",
            ]
        ),
        "properties": {
            "annotated_as": mappings.keyword_field(),
            "predicted_as": mappings.keyword_field(),
            "text_predicted": mappings.text_field(),
            "score": mappings.decimal_field(),
        },
    }
